import sys

import threading
import torch.multiprocessing as multiprocessing
from torch.utils.data import DataLoader as TorchDataLoader
import itertools
import torch
import random
from torch._six import queue
from torch._utils import ExceptionWrapper
from torch.utils.data._utils import signal_handling, MP_STATUS_CHECK_INTERVAL, pin_memory
from torch.utils.data._utils.worker import ManagerWatchdog, WorkerInfo, _IterableDatasetStopIteration
import torch.utils.data._utils as _utils
from torch.utils.data._utils.collate import default_collate

from importlib import import_module

# dataset工厂 工厂模式


def worker_loop(dataset_kind, dataset, index_queue, data_queue, done_event,
                 auto_collation, collate_fn,  drop_last, seed, init_fn, worker_id,
                 num_workers):
    # See NOTE [ Data Loader Multiprocessing Shutdown Logic ] for details on the
    # logic of this function.

    try:
        # Intialize C side signal handlers for SIGBUS and SIGSEGV. Python signal
        # module's handlers are executed after Python returns from C low-level
        # handlers, likely when the same fatal signal had already happened
        # again.
        # https://docs.python.org/3/library/signal.html#execution-of-python-signal-handlers
        signal_handling._set_worker_signal_handlers()

        torch.set_num_threads(1)
        random.seed(seed)
        torch.manual_seed(seed)

        global _worker_info
        _worker_info = WorkerInfo(id=worker_id, num_workers=num_workers,
                                  seed=seed, dataset=dataset)

        from torch.utils.data import _DatasetKind

        init_exception = None

        try:
            if init_fn is not None:
                init_fn(worker_id)

            fetcher = _DatasetKind.create_fetcher(dataset_kind, dataset, auto_collation, collate_fn, drop_last)
        except Exception:
            init_exception = ExceptionWrapper(
                where="in DataLoader worker process {}".format(worker_id))

        # When using Iterable mode, some worker can exit earlier than others due
        # to the IterableDataset behaving differently for different workers.
        # When such things happen, an `_IterableDatasetStopIteration` object is
        # sent over to the main process with the ID of this worker, so that the
        # main process won't send more tasks to this worker, and will send
        # `None` to this worker to properly exit it.
        #
        # Note that we cannot set `done_event` from a worker as it is shared
        # among all processes. Instead, we set the `iteration_end` flag to
        # signify that the iterator is exhausted. When either `done_event` or
        # `iteration_end` is set, we skip all processing step and just wait for
        # `None`.
        iteration_end = False

        watchdog = ManagerWatchdog()



        while watchdog.is_alive():
            try:
                r = index_queue.get(timeout=MP_STATUS_CHECK_INTERVAL)
            except queue.Empty:
                continue
            if r is None:
                # Received the final signal
                assert done_event.is_set() or iteration_end
                break
            elif done_event.is_set() or iteration_end:
                # `done_event` is set. But I haven't received the final signal
                # (None) yet. I will keep continuing until get it, and skip the
                # processing steps.
                continue
            idx, index = r
            if init_exception is not None:
                data = init_exception
                init_exception = None
            else:
                try:
                    idx_scale = 0
                    scale_list = dataset.scale;
                    if len(scale_list) > 0:
                        idx_scale = random.randrange(0, len(scale_list))
                        dataset.set_scale(idx_scale)

                    data = fetcher.fetch(index)
                    data.append(idx_scale)

                except Exception as e:
                    if isinstance(e, StopIteration) and dataset_kind == _DatasetKind.Iterable:
                        data = _IterableDatasetStopIteration(worker_id)
                        # Set `iteration_end`
                        #   (1) to save future `next(...)` calls, and
                        #   (2) to avoid sending multiple `_IterableDatasetStopIteration`s.
                        iteration_end = True
                    else:
                        # It is important that we don't store exc_info in a variable.
                        # `ExceptionWrapper` does the correct thing.
                        # See NOTE [ Python Traceback Reference Cycle Problem ]
                        data = ExceptionWrapper(
                            where="in DataLoader worker process {}".format(worker_id))
            data_queue.put((idx, data))
            del data, idx, index, r  # save memory
    except KeyboardInterrupt:
        # Main process will raise KeyboardInterrupt anyways.
        pass
    if done_event.is_set():
        data_queue.cancel_join_thread()
        data_queue.close()

class _DatasetKind(object):
    Map = 0
    Iterable = 1

    @staticmethod
    def create_fetcher(kind, dataset, auto_collation, collate_fn, drop_last):
        if kind == _DatasetKind.Map:
            return _utils.fetch._MapDatasetFetcher(dataset, auto_collation, collate_fn, drop_last)
        else:
            return _utils.fetch._IterableDatasetFetcher(dataset, auto_collation, collate_fn, drop_last)

class _BaseDataLoaderIter(object):
    def __init__(self, loader):
        self.dataset = loader.dataset
        self.dataset_kind = loader._dataset_kind
        self.auto_collation = loader._auto_collation
        self.drop_last = loader.drop_last
        self.index_sampler = loader._index_sampler
        self.num_workers = loader.num_workers
        self.pin_memory = loader.pin_memory and torch.cuda.is_available()
        self.timeout = loader.timeout
        self.collate_fn = loader.collate_fn
        self.sampler_iter = iter(self.index_sampler)
        self.base_seed = torch.empty((), dtype=torch.int64).random_().item()

    def __iter__(self):
        return self

    def _next_index(self):
        return next(self.sampler_iter)  # may raise StopIteration

    def __next__(self):
        raise NotImplementedError

    def __len__(self):
        return len(self.index_sampler)

    def __getstate__(self):
        # TODO: add limited pickling support for sharing an iterator
        # across multiple threads for HOGWILD.
        # Probably the best way to do this is by moving the sample pushing
        # to a separate thread and then just sharing the data queue
        # but signalling the end is tricky without a non-blocking API
        raise NotImplementedError("{} cannot be pickled", self.__class__.__name__)

        # self._try_put_index()

class MultiProcDataLoaderIter(_BaseDataLoaderIter):
    def __init__(self, loader):
        super(MultiProcDataLoaderIter, self).__init__(loader)

        assert self.num_workers > 0

        if loader.multiprocessing_context is None:
            multiprocessing_context = multiprocessing
        else:
            multiprocessing_context = loader.multiprocessing_context

        self.worker_init_fn = loader.worker_init_fn
        self.worker_queue_idx_cycle = itertools.cycle(range(self.num_workers))
        self.worker_result_queue = multiprocessing_context.Queue()
        self.worker_pids_set = False
        self.shutdown = False
        self.send_idx = 0  # idx of the next task to be sent to workers
        self.rcvd_idx = 0  # idx of the next task to be returned in __next__
        # information about data not yet yielded, i.e., tasks w/ indices in range [rcvd_idx, send_idx).
        # map: task idx => - (worker_id,)        if data isn't fetched (outstanding)
        #                  \ (worker_id, data)   if data is already fetched (out-of-order)
        self.task_info = {}
        self.tasks_outstanding = 0  # always equal to count(v for v in task_info.values() if len(v) == 1)
        self.workers_done_event = multiprocessing_context.Event()

        self.index_queues = []
        self.workers = []
        # A list of booleans representing whether each worker still has work to
        # do, i.e., not having exhausted its iterable dataset object. It always
        # contains all `True`s if not using an iterable-style dataset
        # (i.e., if kind != Iterable).
        self.workers_status = []
        for i in range(self.num_workers):
            index_queue = multiprocessing_context.Queue()
            # index_queue.cancel_join_thread()
            w = multiprocessing_context.Process(
                target=worker_loop,
                args=(self.dataset_kind, self.dataset, index_queue,
                      self.worker_result_queue, self.workers_done_event,
                      self.auto_collation, self.collate_fn, self.drop_last,
                      self.base_seed + i, self.worker_init_fn, i, self.num_workers))
            w.daemon = True
            # NB: Process.start() actually take some time as it needs to
            #     start a process and pass the arguments over via a pipe.
            #     Therefore, we only add a worker to self.workers list after
            #     it started, so that we do not call .join() if program dies
            #     before it starts, and __del__ tries to join but will get:
            #     AssertionError: can only join a started process.
            w.start()
            self.index_queues.append(index_queue)
            self.workers.append(w)
            self.workers_status.append(True)

        if self.pin_memory:
            self.pin_memory_thread_done_event = threading.Event()
            self.data_queue = queue.Queue()
            pin_memory_thread = threading.Thread(
                target=pin_memory._pin_memory_loop,
                args=(self.worker_result_queue, self.data_queue,
                      torch.cuda.current_device(),
                      self.pin_memory_thread_done_event))
            pin_memory_thread.daemon = True
            pin_memory_thread.start()
            # Similar to workers (see comment above), we only register
            # pin_memory_thread once it is started.
            self.pin_memory_thread = pin_memory_thread
        else:
            self.data_queue = self.worker_result_queue

        signal_handling._set_worker_pids(id(self), tuple(w.pid for w in self.workers))
        signal_handling._set_SIGCHLD_handler()
        self.worker_pids_set = True

        # prime the prefetch loop
        for _ in range(2 * self.num_workers):
            self._try_put_index()

    def _try_get_data(self, timeout=MP_STATUS_CHECK_INTERVAL):
        # Tries to fetch data from `self.data_queue` once for a given timeout.
        # This can also be used as inner loop of fetching without timeout, with
        # the sender status as the loop condition.
        #
        # This raises a `RuntimeError` if any worker died expectedly. This error
        # can come from either the SIGCHLD handler in `_utils/signal_handling.py`
        # (only for non-Windows platforms), or the manual check below on errors
        # and timeouts.
        #
        # Returns a 2-tuple:
        #   (bool: whether successfully get data, any: data if successful else None)
        try:
            data = self.data_queue.get(timeout=timeout)
            return (True, data)
        except Exception as e:
            # At timeout and error, we manually check whether any worker has
            # failed. Note that this is the only mechanism for Windows to detect
            # worker failures.
            failed_workers = []
            for worker_id, w in enumerate(self.workers):
                if self.workers_status[worker_id] and not w.is_alive():
                    failed_workers.append(w)
                    self._shutdown_worker(worker_id)
            if len(failed_workers) > 0:
                pids_str = ', '.join(str(w.pid) for w in failed_workers)
                raise RuntimeError('DataLoader worker (pid(s) {}) exited unexpectedly'.format(pids_str))
            if isinstance(e, queue.Empty):
                return (False, None)
            raise

    def _get_data(self):
        # Fetches data from `self.data_queue`.
        #
        # We check workers' status every `MP_STATUS_CHECK_INTERVAL` seconds,
        # which we achieve by running `self._try_get_data(timeout=MP_STATUS_CHECK_INTERVAL)`
        # in a loop. This is the only mechanism to detect worker failures for
        # Windows. For other platforms, a SIGCHLD handler is also used for
        # worker failure detection.
        #
        # If `pin_memory=True`, we also need check if `pin_memory_thread` had
        # died at timeouts.
        if self.timeout > 0:
            success, data = self._try_get_data(self.timeout)
            if success:
                return data
            else:
                raise RuntimeError('DataLoader timed out after {} seconds'.format(self.timeout))
        elif self.pin_memory:
            while self.pin_memory_thread.is_alive():
                success, data = self._try_get_data()
                if success:
                    return data
            else:
                # while condition is false, i.e., pin_memory_thread died.
                raise RuntimeError('Pin memory thread exited unexpectedly')
            # In this case, `self.data_queue` is a `queue.Queue`,. But we don't
            # need to call `.task_done()` because we don't use `.join()`.
        else:
            while True:
                success, data = self._try_get_data()
                if success:
                    return data

    def __next__(self):
        while True:
            # If the worker responsible for `self.rcvd_idx` has already ended
            # and was unable to fulfill this task (due to exhausting an `IterableDataset`),
            # we try to advance `self.rcvd_idx` to find the next valid index.
            #
            # This part needs to run in the loop because both the `self._get_data()`
            # call and `_IterableDatasetStopIteration` check below can mark
            # extra worker(s) as dead.
            while self.rcvd_idx < self.send_idx:
                info = self.task_info[self.rcvd_idx]
                worker_id = info[0]
                if len(info) == 2 or self.workers_status[worker_id]:  # has data or is still active
                    break
                del self.task_info[self.rcvd_idx]
                self.rcvd_idx += 1
            else:
                # no valid `self.rcvd_idx` is found (i.e., didn't break)
                self._shutdown_workers()
                raise StopIteration

            # Now `self.rcvd_idx` is the batch index we want to fetch

            # Check if the next sample has already been generated
            if len(self.task_info[self.rcvd_idx]) == 2:
                data = self.task_info.pop(self.rcvd_idx)[1]
                return self._process_data(data)

            assert not self.shutdown and self.tasks_outstanding > 0
            idx, data = self._get_data()
            self.tasks_outstanding -= 1

            if self.dataset_kind == _DatasetKind.Iterable:
                # Check for _IterableDatasetStopIteration
                if isinstance(data, _utils.worker._IterableDatasetStopIteration):
                    self._shutdown_worker(data.worker_id)
                    self._try_put_index()
                    continue

            if idx != self.rcvd_idx:
                # store out-of-order samples
                self.task_info[idx] += (data,)
            else:
                del self.task_info[idx]
                return self._process_data(data)

    next = __next__  # Python 2 compatibility

    def _try_put_index(self):
        assert self.tasks_outstanding < 2 * self.num_workers
        try:
            index = self._next_index()
        except StopIteration:
            return
        for _ in range(self.num_workers):  # find the next active worker, if any
            worker_queue_idx = next(self.worker_queue_idx_cycle)
            if self.workers_status[worker_queue_idx]:
                break
        else:
            # not found (i.e., didn't break)
            return

        self.index_queues[worker_queue_idx].put((self.send_idx, index))
        self.task_info[self.send_idx] = (worker_queue_idx,)
        self.tasks_outstanding += 1
        self.send_idx += 1

    def _process_data(self, data):
        self.rcvd_idx += 1
        self._try_put_index()
        if isinstance(data, ExceptionWrapper):
            data.reraise()
        return data

    def _shutdown_worker(self, worker_id):
        # Mark a worker as having finished its work and dead, e.g., due to
        # exhausting an `IterableDataset`. This should be used only when this
        # `_MultiProcessingDataLoaderIter` is going to continue running.

        assert self.workers_status[worker_id]

        # Signal termination to that specific worker.
        q = self.index_queues[worker_id]
        # Indicate that no more data will be put on this queue by the current
        # process.
        q.put(None)

        # Note that we don't actually join the worker here, nor do we remove the
        # worker's pid from C side struct because (1) joining may be slow, and
        # (2) since we don't join, the worker may still raise error, and we
        # prefer capturing those, rather than ignoring them, even though they
        # are raised after the worker has finished its job.
        # Joinning is deferred to `_shutdown_workers`, which it is called when
        # all workers finish their jobs (e.g., `IterableDataset` replicas) or
        # when this iterator is garbage collected.
        self.workers_status[worker_id] = False

    def _shutdown_workers(self):
        # Called when shutting down this `_MultiProcessingDataLoaderIter`.
        # See NOTE [ Data Loader Multiprocessing Shutdown Logic ] for details on
        # the logic of this function.
        python_exit_status = _utils.python_exit_status
        if python_exit_status is True or python_exit_status is None:
            # See (2) of the note. If Python is shutting down, do no-op.
            return
        # Normal exit when last reference is gone / iterator is depleted.
        # See (1) and the second half of the note.
        if not self.shutdown:
            self.shutdown = True
            try:
                # Exit `pin_memory_thread` first because exiting workers may leave
                # corrupted data in `worker_result_queue` which `pin_memory_thread`
                # reads from.
                if hasattr(self, 'pin_memory_thread'):
                    # Use hasattr in case error happens before we set the attribute.
                    self.pin_memory_thread_done_event.set()
                    # Send something to pin_memory_thread in case it is waiting
                    # so that it can wake up and check `pin_memory_thread_done_event`
                    self.worker_result_queue.put((None, None))
                    self.pin_memory_thread.join()
                    self.worker_result_queue.close()

                # Exit workers now.
                self.workers_done_event.set()
                for worker_id in range(len(self.workers)):
                    # Get number of workers from `len(self.workers)` instead of
                    # `self.num_workers` in case we error before starting all
                    # workers.
                    if self.workers_status[worker_id]:
                        self._shutdown_worker(worker_id)
                for w in self.workers:
                    w.join()
                for q in self.index_queues:
                    q.cancel_join_thread()
                    q.close()
            finally:
                # Even though all this function does is putting into queues that
                # we have called `cancel_join_thread` on, weird things can
                # happen when a worker is killed by a signal, e.g., hanging in
                # `Event.set()`. So we need to guard this with SIGCHLD handler,
                # and remove pids from the C side data structure only at the
                # end.
                #
                # FIXME: Unfortunately, for Windows, we are missing a worker
                #        error detection mechanism here in this function, as it
                #        doesn't provide a SIGCHLD handler.
                if self.worker_pids_set:
                    signal_handling._remove_worker_pids(id(self))
                    self.worker_pids_set = False

    def __del__(self):
        self._shutdown_workers()

class SingleProcDataLoaderIter(_BaseDataLoaderIter):
    def __init__(self, loader):
        super(SingleProcDataLoaderIter, self).__init__(loader)
        assert self.timeout == 0
        assert self.num_workers == 0

        self.dataset_fetcher = _DatasetKind.create_fetcher(
            self.dataset_kind, self.dataset, self.auto_collation, self.collate_fn, self.drop_last)

    def __next__(self):
        index = self._next_index()  # may raise StopIteration
        data = self.dataset_fetcher.fetch(index)  # may raise StopIteration
        if self.pin_memory:
            data = _utils.pin_memory.pin_memory(data)
        return data

    next = __next__  # Python 2 compatibility



class MSDataLoader(TorchDataLoader):
    def __init__(self, args, dataset, batch_size=1, shuffle=False, num_workers=0,
        sampler=None, batch_sampler=None,
        collate_fn=default_collate, pin_memory=False, drop_last=False,
        timeout=0, worker_init_fn=None):

        super(MSDataLoader, self).__init__(
            dataset, batch_size=batch_size, shuffle=shuffle,
            sampler=sampler, batch_sampler=batch_sampler,
            num_workers=num_workers, collate_fn=collate_fn,
            pin_memory=pin_memory, drop_last=drop_last,
            timeout=timeout, worker_init_fn=worker_init_fn)
        # super(MSDataLoader, self).__init__(dataset=dataset, batch_size=args.batch_size, shuffle=args.shuffle, sampler=None,
        #          batch_sampler=None, num_workers=args.num_workers, collate_fn=None,
        #          pin_memory=args.pin_memory, drop_last=False, timeout=0,
        #          worker_init_fn=None, multiprocessing_context=None)

        self.args = args
        # self._sr_scales = dataset.sr_scales
        # self._sr_task = dataset.sr_task

    def __iter__(self):
        if self.num_workers == 0:
            return SingleProcDataLoaderIter(self)
        else:
            return MultiProcDataLoaderIter(self)
