import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from math import floor
###NL-block###############
class _NonLocalBlockND(nn.Module):
    def __init__(self, in_channels, inter_channels=None, dimension=3, sub_sample=True, bn_layer=True):
        super(_NonLocalBlockND, self).__init__()

        assert dimension in [1, 2, 3]

        self.dimension = dimension
        self.sub_sample = sub_sample

        self.in_channels = in_channels
        self.inter_channels = inter_channels

        if self.inter_channels is None:
            self.inter_channels = in_channels//2
            if self.inter_channels == 0:
                self.inter_channels = 1

        if dimension == 3:
            conv_nd = nn.Conv3d
            max_pool_layer = nn.MaxPool3d(kernel_size=(1, 2, 2))
            bn = nn.BatchNorm3d
        elif dimension == 2:
            conv_nd = nn.Conv2d
            max_pool_layer = nn.MaxPool2d(kernel_size=(2, 2))
            bn = nn.BatchNorm2d
        else:
            conv_nd = nn.Conv1d
            max_pool_layer = nn.MaxPool1d(kernel_size=(2))
            bn = nn.BatchNorm1d

        self.g = conv_nd(in_channels=self.in_channels, out_channels=self.inter_channels,
                         kernel_size=1, stride=1, padding=0)

        if bn_layer:
            self.W = nn.Sequential(
                conv_nd(in_channels=self.inter_channels, out_channels=self.in_channels,
                        kernel_size=1, stride=1, padding=0),
                bn(self.in_channels)
            )
            nn.init.constant_(self.W[1].weight, 0)
            nn.init.constant_(self.W[1].bias, 0)
        else:
            self.W = conv_nd(in_channels=self.inter_channels, out_channels=self.in_channels,
                             kernel_size=1, stride=1, padding=0)
            nn.init.constant_(self.W.weight, 0)
            nn.init.constant_(self.W.bias, 0)

        self.theta = conv_nd(in_channels=self.in_channels, out_channels=self.inter_channels,
                             kernel_size=1, stride=1, padding=0)
        self.phi = conv_nd(in_channels=self.in_channels, out_channels=self.inter_channels,
                           kernel_size=1, stride=1, padding=0)

        if sub_sample:
            self.g = nn.Sequential(self.g, max_pool_layer)
            self.phi = nn.Sequential(self.phi, max_pool_layer)

    def forward(self, x):
        '''
        :param x: (b, c, t, h, w)
        :return:
        '''

        batch_size = x.size(0)

        g_x = self.g(x).view(batch_size, self.inter_channels, -1)
        g_x = g_x.permute(0, 2, 1)#矩阵维度换位

        theta_x = self.theta(x).view(batch_size, self.inter_channels, -1)
        theta_x = theta_x.permute(0, 2, 1)
        phi_x = self.phi(x).view(batch_size, self.inter_channels, -1)
        f = torch.matmul(theta_x, phi_x)#矩阵相乘
        f_div_C = F.softmax(f, dim=-1)

        y = torch.matmul(f_div_C, g_x)
        y = y.permute(0, 2, 1).contiguous()#转化为连续
        y = y.view(batch_size, self.inter_channels, *x.size()[2:])
        W_y = self.W(y)
        z = W_y + x

        return z


class NONLocalBlock1D(_NonLocalBlockND):
    def __init__(self, in_channels, inter_channels=None, sub_sample=True, bn_layer=True):
        super(NONLocalBlock1D, self).__init__(in_channels,
                                              inter_channels=inter_channels,
                                              dimension=1, sub_sample=sub_sample,
                                              bn_layer=bn_layer)


class NONLocalBlock2D(_NonLocalBlockND):
    def __init__(self, in_channels, inter_channels=None, sub_sample=True, bn_layer=True):
        super(NONLocalBlock2D, self).__init__(in_channels,
                                              inter_channels=inter_channels,
                                              dimension=2, sub_sample=sub_sample,
                                              bn_layer=bn_layer)


class NONLocalBlock3D(_NonLocalBlockND):
    def __init__(self, in_channels, inter_channels=None, sub_sample=True, bn_layer=True):
        super(NONLocalBlock3D, self).__init__(in_channels,
                                              inter_channels=inter_channels,
                                              dimension=3, sub_sample=sub_sample,
                                              bn_layer=bn_layer)


def default_conv(in_channels, out_channels, kernel_size, bias=True):
    return nn.Conv2d(
        in_channels, out_channels, kernel_size,
        padding=(kernel_size//2), bias=bias)

class MeanShift(nn.Conv2d):
    def __init__(self, rgb_range, rgb_mean, rgb_std, sign=-1):
        super(MeanShift, self).__init__(3, 3, kernel_size=1)
        std = torch.Tensor(rgb_std)
        self.weight.data = torch.eye(3).view(3, 3, 1, 1)
        self.weight.data.div_(std.view(3, 1, 1, 1))
        self.bias.data = sign * rgb_range * torch.Tensor(rgb_mean)
        self.bias.data.div_(std)
        self.requires_grad = False

class BasicBlock(nn.Sequential):
    def __init__(
        self, in_channels, out_channels, kernel_size, stride=1, bias=False,
        bn=True, act=nn.ReLU(True)):

        m = [nn.Conv2d(
            in_channels, out_channels, kernel_size,
            padding=(kernel_size//2), stride=stride, bias=bias)
        ]
        if bn: m.append(nn.BatchNorm2d(out_channels))
        if act is not None: m.append(act)
        super(BasicBlock, self).__init__(*m)

class ResBlock(nn.Module):
    def __init__(
        self, conv, n_feat, kernel_size,
        bias=True, bn=False, act=nn.ReLU(True), res_scale=1):

        super(ResBlock, self).__init__()
        m = []

        for i in range(2):
            m.append(conv(n_feat, n_feat, kernel_size, bias=bias))
            if bn: m.append(nn.BatchNorm2d(n_feat))
            if i == 0: m.append(act)

        self.body = nn.Sequential(*m)
        self.res_scale = res_scale

    def forward(self, x):
        res = self.body(x).mul(self.res_scale)
        res += x

        return res

# class Upsampler(nn.Sequential):
#     def __init__(self, conv, scale, n_feat, bn=False, act=False, bias=True):
#
#         m = []
#         if (scale & (scale - 1)) == 0:    # Is scale = 2^n?
#             for _ in range(int(math.log(scale, 2))):
#                 m.append(conv(n_feat, 4 * n_feat, 3, bias))
#                 m.append(nn.PixelShuffle(2))
#                 if bn: m.append(nn.BatchNorm2d(n_feat))
#                 if act: m.append(act())
#         elif scale == 3:
#             m.append(conv(n_feat, 9 * n_feat, 3, bias))
#             m.append(nn.PixelShuffle(3))
#             if bn: m.append(nn.BatchNorm2d(n_feat))
#             if act: m.append(act())
#         else:
#             raise NotImplementedError
#
#         super(Upsampler, self).__init__(*m)

def psf2otf(psf, shape):
    eps = torch.finfo(psf.dtype).eps
    psf_shape = torch.tensor(psf.shape)
    otf = torch.zeros(psf.shape[:-2] + shape).type_as(psf)
    otf[...,:psf.shape[2],:psf.shape[3]].copy_(psf)
    ndim = psf_shape.shape[0]
    # print(ndim)
    for axis, axis_size in enumerate(psf.shape[2:]):
        otf = torch.roll(otf, -int(floor(axis_size / 2)), dims=axis + 2)

    otf = torch.fft.fft2(otf, dim=(-2, -1))


    otf_real = torch.real(otf)
    otf_img = torch.imag(otf)

    n_ops = 0
    n_elem = torch.prod(psf_shape[2: ])
    for k in range(2, ndim):
        nffts = n_elem / psf_shape[k]
        n_ops = n_ops + psf.shape[k] * torch.log2(psf_shape[k].float()) * nffts

    max_val, idx = torch.max(torch.abs(otf_img.flatten(2)), dim=2, keepdim=True)
    max_abs_val, abs_idx = torch.max(torch.abs(otf.flatten(2)), dim=2, keepdim=True)
    rate_map = (max_val / max_abs_val) <= n_ops * eps# 2.2204e-16

    for i in range(psf_shape[0].item()):
        for j in range(psf_shape[1].item()):
            if rate_map[i, j] == True:
                otf_img[i, j, :, :] = 0

    otf = torch.complex(otf_real, otf_img)

    return otf

class L0Smoothing(nn.Module):
    def __init__(self, in_channels, lambda_=0.0025, kappa=2.0):  #lambda_=2e-2  kappa=2.0
        super(L0Smoothing, self).__init__()
        self.fx = nn.Parameter(torch.tensor([[[[1, -1]]]], dtype=torch.float32), requires_grad=False)
        self.fy = nn.Parameter(torch.tensor([[[[1], [-1]]]], dtype=torch.float32), requires_grad=False)

        self.betamax = 1e5
        # self.padx = partial(F.pad, mode=self.pad_mode, pad=(0, 1, 0, 0))
        # self.pady = partial(F.pad, mode=self.pad_mode, pad=(0, 0, 0, 1))
        self.lambda_ = lambda_
        self.kappa = kappa
        self.in_channels = in_channels

    def forward(self, image):
        S = image
        B, _, N, M = image.shape
        sizeI2D = (N, M)
        otfFx = psf2otf(self.fx, sizeI2D)
        otfFy = psf2otf(self.fy, sizeI2D)
        Normin1 = torch.fft.fft2(image)
        Denormin2 = torch.abs(otfFx).pow(2) + torch.abs(otfFy).pow(2)

        if self.in_channels > 1:
            Denormin2 = Denormin2.repeat([1, self.in_channels, 1, 1])

        beta = 2 * self.lambda_

        while beta < self.betamax:
            Denormin = 1 + beta * Denormin2
            h = torch.cat(
                [
                    S[:, :, :, 1:] - S[:, :, :, 0:M-1],
                    (S[:, :, :, 0] - S[:, :, :, -1]).unsqueeze(3)
                ],
                dim=3)

            v = torch.cat(
                [
                    S[:, :, 1:, :] - S[:, :, 0: N - 1, :],
                    (S[:, :, 0, :] - S[:, :, -1, :]).unsqueeze(2)
                ],
                dim=2)

            # idx_mask = t.sum(h.pow(2) + v.pow(2), dim=1, keepdim=True)
            if self.in_channels == 1:
                idx_mask = (h.pow(2) + v.pow(2)) < (self.lambda_ / beta)
            else:
                idx_mask = torch.sum(h.pow(2) + v.pow(2), dim=1, keepdim=True)
                idx_mask = idx_mask < (self.lambda_ / beta)
                idx_mask = idx_mask.repeat([1, self.in_channels, 1, 1])

            h[idx_mask] = 0
            v[idx_mask] = 0

            Normin2 = torch.cat(
                [
                    (h[:, :, :, -1] - h[:, :, :, 0]).unsqueeze(3),
                    h[:, :, :, 0:M - 1] - h[:, :, :, 1:],
                ],
                dim=3
            )

            # F.conv2d(self.padx(h),  self.fx, stride=1, groups=self.in_channels)
            Normin2 = Normin2 + torch.cat(
                [
                    (v[:, :, -1, :] - v[:, :, 0, :]).unsqueeze(2),
                    v[:, :, 0: N - 1, :] - v[:, :, 1:, :],
                ],
                dim=2
            )

            Fs = (Normin1 + beta * torch.fft.fft2(Normin2)) / Denormin.cuda()

            S = torch.real(torch.fft.ifft2(Fs))
            beta *= self.kappa


        return S