import os
import math
from decimal import Decimal

import utility
from loss import losses

import torch
from loss.ssim_loss import MS_SSIM
from tqdm import tqdm

class Trainer():
    def __init__(self, args, loader, my_model, my_loss, ckp):
        self.args = args
        self.scale = args.scale

        self.ckp = ckp
        self.loader_train = loader.loader_train
        self.loader_test = loader.loader_test
        self.model = my_model
        self.loss = my_loss
        self.optimizer = utility.make_optimizer(args, self.model)
        self.scheduler = utility.make_scheduler(args, self.optimizer)
        self.rgb_range = self.args.rgb_range
        if self.args.load != '.':
            self.optimizer.load_state_dict(
                torch.load(os.path.join(ckp.dir, 'optimizer.pt'))
            )
            for _ in range(len(ckp.log)): self.scheduler.step()

        self.error_last = 1e8


    def train(self):
        self.scheduler.step()
        self.loss.step()
        epoch = self.scheduler.last_epoch
        lr = self.scheduler.get_lr()[0]

        self.ckp.write_log(
            '[Epoch {}]\tLearning rate: {:.2e}'.format(epoch, Decimal(lr))
        )
        self.loss.start_log()
        self.model.train()

        timer_data, timer_model = utility.timer(), utility.timer()
        for batch, (lr, hr, idx_scale) in enumerate(self.loader_train):
            lr, hr = self.prepare([lr, hr])
            timer_data.hold()
            timer_model.tic()

            self.optimizer.zero_grad()
            #
            # sr = self.model(lr, idx_scale)
            # loss = self.loss(sr, hr)

########################L1 loss
            sr, sam = self.model(lr, idx_scale)
            loss_l1 = self.loss(sr, hr)
            #######Charbonnier_loss
            char = losses.CharbonnierLoss()
            Char_loss = char(sam, hr)


            loss = loss_l1 + Char_loss


            if loss.item() < self.args.skip_threshold * self.error_last:
                loss.backward()
                self.optimizer.step()
            else:
                print('Skip this batch {}! (Loss: {})'.format(
                    batch + 1, loss.item()
                ))

            timer_model.hold()

            if (batch + 1) % self.args.print_every == 0:
                self.ckp.write_log('[{}/{}]\t{}\t{:.1f}+{:.1f}s'.format(
                    (batch + 1) * self.args.batch_size,
                    len(self.loader_train.dataset),
                    self.loss.display_loss(batch),
                    timer_model.release(),
                    timer_data.release()))

            timer_data.tic()

        self.loss.end_log(len(self.loader_train))
        self.error_last = self.loss.log[-1, -1]

    def test(self):
        epoch = self.scheduler.last_epoch
        self.ckp.write_log('\nEvaluation:')
        self.ckp.add_log(torch.zeros(1, len(self.scale)))
        self.model.eval()

        timer_test = utility.timer()

        with torch.no_grad():
            for idx_scale, scale in enumerate(self.scale):
                eval_acc = 0
                self.loader_test.dataset.set_scale(idx_scale)
                tqdm_test = tqdm(self.loader_test, ncols=80)
                for idx_img, (lr, hr, filename) in enumerate(tqdm_test):
                    filename = filename[0]
                    no_eval = (hr.nelement() == 1)
                    if not no_eval:
                        lr, hr = self.prepare([lr, hr])
                    else:
                        lr = self.prepare([lr])[0]

                    sr, sam = self.model(lr, idx_scale)
                    # sr = self.model(lr, idx_scale)

                    sr = utility.quantize(sr, self.args.rgb_range)

                    save_list = [sr]
                    if not no_eval:
                        eval_acc += utility.calc_psnr(sr, hr
                            # sr, hr, scale, self.args.rgb_range,
                            # benchmark=self.loader_test.dataset.benchmark
                        )
                        save_list.extend([lr, hr])

                    if self.args.save_results:
                        self.ckp.save_results(filename, save_list, scale)

                self.ckp.log[-1, idx_scale] = eval_acc / len(self.loader_test)
                best = self.ckp.log.max(0)
                self.ckp.write_log(
                    '[{} x{}]\tPSNR: {:.3f} (Best: {:.3f} @epoch {})'.format(
                        self.args.data_test,
                        scale,
                        self.ckp.log[-1, idx_scale],
                        best[0][idx_scale],
                        best[1][idx_scale] + 1
                    )
                )

        self.ckp.write_log(
            'Total time: {:.2f}s\n'.format(timer_test.toc()), refresh=True
        )
        if not self.args.test_only:
            self.ckp.save(self, epoch, is_best=(best[1][0] + 1 == epoch))
    def prepare(self, l, volatile=False):
        device = torch.device('cpu' if self.args.cpu else 'cuda')
        def _prepare(tensor):
            if self.args.precision == 'half': tensor = tensor.half()
            return tensor.to(device)
           
        return [_prepare(_l) for _l in l]

    def terminate(self):
        if self.args.test_only:
            self.test()
            return True
        else:
            epoch = self.scheduler.last_epoch + 1
            return epoch >= self.args.epochs

    # def get_feature1(args, tensor_data, batch_idx, epoch):
    #     args.validate_dir = 'G:/images/res'
    #     b, c, h, w = tensor_data.size()
    #     for _batch in range(b):
    #         for _channel in range(c):
    #             feature = tensor_data[_batch, _channel, :, :]
    #             feature = feature.view(h, w) / 255.0
    #             feature = feature.data.to('cpu').numpy()
    #             # print(feature)
    #             feature = 1.0 / (1 + np.exp(-1 * feature))
    #             feature = np.round(feature * 255)
    #             if not os.path.exists(args.validate_dir):
    #                 os.makedirs(args.validate_dir)
    #             cv2.imwrite(os.path.join(args.validate_dir, 'feat11_%s_%s.jpg' % (str(epoch), str(batch_idx))), feature)
    #
    # def get_feature2(args, tensor_data, batch_idx, epoch):
    #     args.validate_dir = 'G:/images/x_smooth'
    #     b, c, h, w = tensor_data.size()
    #     for _batch in range(b):
    #         for _channel in range(c):
    #             feature = tensor_data[_batch, _channel, :, :]
    #             feature = feature.view(h, w) / 255.0
    #             feature = feature.data.to('cpu').numpy()
    #             # print(feature)
    #             feature = 1.0 / (1 + np.exp(-1 * feature))
    #             feature = np.round(feature * 255)
    #             if not os.path.exists(args.validate_dir):
    #                 os.makedirs(args.validate_dir)
    #             cv2.imwrite(os.path.join(args.validate_dir, 'feat11_%s_%s.jpg' % (str(epoch), str(batch_idx))), feature)
    #
    # def get_feature3(args, tensor_data, batch_idx, epoch):
    #     args.validate_dir = 'G:/images/x_LH_d'
    #     b, c, h, w = tensor_data.size()
    #     for _batch in range(b):
    #         for _channel in range(c):
    #             feature = tensor_data[_batch, _channel, :, :]
    #             feature = feature.view(h, w) / 255.0
    #             feature = feature.data.to('cpu').numpy()
    #             feature = 1.0 / (1 + np.exp(-1 * feature))
    #             feature = np.round(feature * 255)
    #             if not os.path.exists(args.validate_dir):
    #                 os.makedirs(args.validate_dir)
    #             cv2.imwrite(os.path.join(args.validate_dir, 'feat12_%s_%s.jpg' % (str(epoch), str(batch_idx))), feature)
    #
    # def get_feature4(args, tensor_data, batch_idx, epoch):
    #     args.validate_dir = 'G:/images/x_dHH'
    #     b, c, h, w = tensor_data.size()
    #     for _batch in range(b):
    #         for _channel in range(c):
    #             feature = tensor_data[_batch, _channel, :, :]
    #             feature = feature.view(h, w) / 255.0
    #             feature = feature.data.to('cpu').numpy()
    #             feature = 1.0 / (1 + np.exp(-1 * feature))
    #             feature = np.round(feature * 255)
    #             if not os.path.exists(args.validate_dir):
    #                 os.makedirs(args.validate_dir)
    #             cv2.imwrite(os.path.join(args.validate_dir, 'feat21_%s_%s.jpg' % (str(epoch), str(batch_idx))), feature)
    #
